/* 
 * vinl_grammar.g : the lexer and parser in ANTLR grammar for VINL 
 *
 * @author Yang Cao  
 *
 * @version $id:
 */

class VinlLexer extends Lexer;

options {
	k = 2;
	charVocabulary = '\3'..'\377';
	testLiterals = false;
	exportVocab = VinlAntlr;
}

protected 
ALPHA	: 'a'..'z' | 'A'..'Z' | '_';

protected
DIGIT	: '0'..'9';

WS	: (' ' | '\t')+		{$setType(Token.SKIP);};

EOL	: ('\n' | ('\r' '\n') => '\r' '\n' | '\r') {$setType(Token.SKIP); newline();};

COMMENT	: "//" (~('\n' | '\r'))* (EOL) {$setType(Token.SKIP);};

LPAREN	: '(';
RPAREN	: ')';
MULT	: '*';
DIV	: '/';
MOD	: '%';
PLUS	: '+';
MINUS	: '-';
SEMI	: ';';
LBRACE	: '{';
RBRACE	: '}';
ASSIGN	: "<-";
COMMA	: ',';
EQ	: '=';
NEQ	: "!=";
GT	: '>';
GE	: ">=";
LT	: '<';
LE	: "<=";
AND	: "&&";
OR	: "||";

ID	options { testLiterals = true; } 
	: ALPHA (ALPHA|DIGIT)* 
	;

/*
 * DECIMAL: consists of an mandator integer part with an optional fraction part
 * e.g. 1, 2.34, 567 
 */
DECIMAL	: (DIGIT)+ ('.' (DIGIT)+)? ;

/*
 * STRING: a sequence of chars surrounded by double quotes ", 
 * double quote inside a string is escaped by \.  New line is not allowed. 
 */
/* string is not useful in vinl
STRING	: '"'! 
		(~('"' | '\n') | ('\'!'"'))*
	  '"'! 
	;
*/



/***************************************************************
 *  parser
 ***************************************************************/

class VinlParser extends Parser;

options {
	k = 2;
	buildAST = true;
	exportVocab = VinlAntlr;
}

tokens {
	STATEMENT;
	FUNC_CALL;
	EXPR_LIST;
	PARAMS_LIST;
	FUNC_BODY;
}

cmd
	: ( statement | func_def )
	| ( "exit" | "bye" | "quit" )
		{ System.exit(0); }	
	| EOF!
		{ System.exit(0); }	
	;	

/* only used in ast testing */
program
	: ( statement | func_def )* EOF!
		{ #program = #([STATEMENT, "PROGRAM"], program); }	
	;
	
statement
	: selection_stmt //if
	| expression_stmt
	| iteration_stmt //while and for
	| keyword_stmt
	| LBRACE! (statement)* RBRACE!
		{#statement = #([STATEMENT,"STATEMENT"], statement);}
	;

expression_stmt
	: expression SEMI!
	;

selection_stmt
	: "if"^ LPAREN! expression RPAREN! statement
		(options {greedy = true;}: "else"! statement)?
	;

iteration_stmt
	: while_stmt
	| for_stmt
	;

while_stmt
	: "while"^ LPAREN! expression RPAREN! statement 
	;

for_stmt
	: "for"^ LPAREN! expression SEMI! expression SEMI! expression RPAREN! statement
	;

keyword_stmt
	: pendown_stmt
	| penup_stmt
	| home_stmt
	| forward_stmt
	| backward_stmt
	| left_stmt
	| right_stmt
	| clear_stmt
	| break_stmt
	| continue_stmt
	| return_stmt
	| echo_stmt
	| status_stmt
	;

break_stmt
	: "break" SEMI!
	;

continue_stmt
	: "continue" SEMI!
	;

return_stmt
	: "return"^ (expression)? SEMI!
	;

pendown_stmt
	: ("pendown" | "pd") SEMI!
	;

penup_stmt
	: ("penup" | "pu") SEMI!
	;

home_stmt
	: ("home") SEMI!
	;

forward_stmt
	: ("forward"^ | "fw"^) expression SEMI!	
	;

backward_stmt
	: ("backward"^ | "bw"^) expression SEMI!
	;

right_stmt
	: ("right"^ | "rt"^) expression SEMI!
	;

left_stmt
	: ("left"^ | "lf"^) expression SEMI!
	;

echo_stmt
	: ("echo"^) expression SEMI!
	;

clear_stmt
	: ("clear" | "cl") SEMI!
	;

status_stmt
	: ("status") SEMI!
	;

type_specifier 
	: "decimal" 
	| "boolean"
	;

return_type_specifier
	: type_specifier 
	| "void"
	;

func_def
	: "funct"^ ID LPAREN! params_list RPAREN! func_body 
	;

func_body
	: LBRACE! (statement)* RBRACE!
		{#func_body = #([STATEMENT, "FUNC_BODY"], func_body); }
	;

params_list 
	: param (COMMA! param)*
		{#params_list = #([PARAMS_LIST, "PARAMS_LIST"], params_list);}
	;

param	
	: type_specifier ID
	| "void"
	;
	

expression
	: operative_expr //includes primary_expr
	| variable_decl_expr
	| assignment_expr
	;

primary_expr 
	: ID
	| constant_expr
	| func_call_expr
	;

constant_expr 
	: "true"
	| "false"
	| DECIMAL
	;
		
//func_call_stmt
func_call_expr
	: ID LPAREN! expr_list RPAREN!
		{#func_call_expr=#([FUNC_CALL,"FUNC_CALL"], func_call_expr);} 
	;

expr_list 
	: expression (COMMA! expression)*
		{#expr_list = #([EXPR_LIST, "EXPR_LIST"], expr_list);}
	;

variable_decl_expr
	: "var"^ type_specifier ID (ASSIGN^ expression)? 
	| "void" //null variable place holder expression
	;

assignment_expr
	: lvalue ASSIGN^ expression 
	;

operative_expr
	: logical_expr (OR^ logical_expr)* 
	;

logical_expr
	: relational_expr (AND^ relational_expr)* 
	;

//no chaining of relational expressions
relational_expr
	: additive_expr ( (LT^ | LE^ | GT^ | GE^ | EQ^ | NEQ^) additive_expr)? 
	;

additive_expr 
	: multiplicative_expr ( (PLUS^ | MINUS^) multiplicative_expr)* 
	;

multiplicative_expr 
	: rvalue ( (MULT^ | DIV^ | MOD^) rvalue)* 
	;

rvalue
	: primary_expr 
	| LPAREN! expression RPAREN!
	;

lvalue
	: ID 
	;

