/* 
 * vinl_walker.g : the AST walker in ANTLR grammar for VINL 
 *
 * @author Yang Cao  
 *
 * @version $id:
 */

{
	import java.io.*;
	import java.util.*;
	import java.awt.*;
}

class VinlAstWalker extends TreeParser;
options {
	importVocab = VinlAntlr;
}

{
	static Hashtable<String, VinlFunction> function_table = new Hashtable();
	static VinlSymbolTable global_table = new VinlSymbolTable();
	static VinlSymbolTable curr_table = global_table;
	static Hashtable<String, String> keywords = new Hashtable();
	static VinlGraphicPanel gp;
	static VinlTurtle turtle;
	VinlType ret = null; // used in function return
	
	/* control flows */
	boolean inloop = false;
	boolean breaked = false;
	boolean continued = false;
	boolean proceed = true; 
	boolean returned = false;

	/* DEBUG LEVEL */
	public static final int INFO = 3;
	public static final int DEBUG = 2;
	public static final int DEBUG_FINE = 1;

	public static int curr_level = 4; //no messages

	static {
		/* vinl keywords should not be used as identifiers.
		   Should also forbid java keywords, however java keywords are not VINL limitation,
		   so I will left it for Java compiler to catch such error.
		*/
		keywords.put("backward", "backward");	
		keywords.put("bw", "bw");	
		keywords.put("boolean", "boolean");	
		keywords.put("break", "break");	
		keywords.put("cl", "cl");	
		keywords.put("clear", "clear");	
		keywords.put("continue", "continue");	
		keywords.put("decimal", "decimal");	
		keywords.put("echo", "echo");	
		keywords.put("for", "for");	
		keywords.put("forward", "forward");	
		keywords.put("funct", "funct");	
		keywords.put("fw", "fw");	
		keywords.put("home", "home");	
		keywords.put("if", "if");	
		keywords.put("left", "left");	
		keywords.put("lt", "lt");	
		keywords.put("pendown", "pendown");	
		keywords.put("pd", "pd");	
		keywords.put("penup", "penup");	
		keywords.put("pu", "pu");	
		keywords.put("right", "right");	
		keywords.put("rt", "rt");	
		keywords.put("var", "var");	
		keywords.put("void", "void");	
		
		gp = VinlInterpreterGui.getGraphicPanel();	
	}
	
	boolean isKeyword(String name) {
		return keywords.containsKey(name);
	}	

	static void log(String s, int level)  {
		if(level > curr_level) 
		{		
			System.out.println(s);
		}
	}
}

expr returns [ VinlType r ] 
{
	VinlType a, b;
	r=null;
	turtle = gp.getTurtle();
	String[] func_params;
	VinlType[] expr_list;
}
	: #(STATEMENT (stmt:. { r=expr(#stmt); })* )
	| #(PLUS a=expr b=expr)			{ if(proceed) r=a.eval(VinlOperation.PLUS, a, b); }
	| #(MINUS a=expr b=expr)		{ if(proceed) r=a.eval(VinlOperation.MINUS, a, b); }
	| #(MULT a=expr b=expr)			{ if(proceed) r=a.eval(VinlOperation.MULTI, a, b); }
	| #(DIV a=expr b=expr)			{ if(proceed) r=a.eval(VinlOperation.DIV, a, b); }
	| #(MOD a=expr b=expr)			{ if(proceed) r=a.eval(VinlOperation.MOD, a, b); }
	| #(LT a=expr b=expr)			{ if(proceed) r=a.eval(VinlOperation.LT, a, b); }
	| #(GT a=expr b=expr)			{ if(proceed) r=a.eval(VinlOperation.GT, a, b); }
	| #(GE a=expr b=expr)			{ if(proceed) r=a.eval(VinlOperation.GE, a, b); }
	| #(EQ a=expr b=expr)			{ if(proceed) r=a.eval(VinlOperation.EQ, a, b); }
	| #(NEQ a=expr b=expr)			{ if(proceed) r=a.eval(VinlOperation.NEQ, a, b); }
	| #(OR a=expr right_or:.)			
		{ 	/* short circuit or */
			if(proceed) 
			{
				if (a instanceof VinlBoolean)
					r=((VinlBoolean) a).getValue()?new VinlBoolean(true):expr(#right_or);
			}
		}
	| #(AND a=expr right_and:.)
		{
			/* short circuit and */
			if(proceed) 
			{
				if (a instanceof VinlBoolean)
					r=((VinlBoolean) a).getValue()?expr(#right_and):new VinlBoolean(false);
			}
		}
	| #(ASSIGN a=expr b=expr)		
		{ 
			if(proceed) 
			{
				if(curr_table.get(a.getName())==null) 
				{
					throw new RecognitionException("ERROR: Could not find variable: "+a.getName());
				} else {
					if(a instanceof VinlBoolean && b instanceof VinlBoolean) 
					{
						((VinlBoolean) a).setValue(((VinlBoolean) b).getValue());
					}
					else if (a instanceof VinlDecimal && b instanceof VinlDecimal) 
					{
						((VinlDecimal) a).setValue(((VinlDecimal) b).getValue());
					} 
					else 
					{
						log("Type of variable: "+((a==null)?"null":a.getType())+ ", type of rvalue: "+((b==null)?"null":b.getType()), INFO);
						throw new RecognitionException("ERROR: Type mismatched assignment.");		
					}
				}
			}
		} 
	| "true"				{ if(proceed) r=new VinlBoolean(true); }
	| "false"				{ if(proceed) r=new VinlBoolean(false); }
	| "void"				{ if(proceed) r = new VinlVoid(); }	
	| "clear"				{ if(proceed) gp.clearScreen(); }
	| "cl"					{ if(proceed) gp.clearScreen(); }
	| "home"				{ if(proceed) gp.homeTurtle(); }
	| "penup"				{ if(proceed) turtle.setPenDown(false); }
	| "pu"					{ if(proceed) turtle.setPenDown(false); }
	| "pendown"				{ if(proceed) turtle.setPenDown(true); }
	| "pd"					{ if(proceed) turtle.setPenDown(true); }
	| "break"				
		{ 
			if(proceed) 
			{ 
				if(inloop) 
				{	
					proceed = false; 
					breaked = true; 
					log("break further stmts", DEBUG);
				} 
				else 
					throw new RecognitionException("ERROR: Illegal break out side of any while/for loop");
			} 
		}
	| "continue"				
		{ 
			if(proceed) 
			{	
				if(inloop) 
				{ 
					proceed = false; 
					continued = true; 
					log("Stop further stmts", DEBUG);
				}
				else 
					throw new RecognitionException("ERROR: Illegal break out side of any while/for loop");
			} 
		}
	| #("var" type:.  id:ID) 			
		{
			if(proceed) 
			{
				if(isKeyword(id.getText())) 
					throw new RecognitionException("ERROR: Illegal use of keyword *"+id.getText()+"*");

				if(type.getText().equals("decimal")) 
				{
					r = new VinlDecimal(id.getText());
					curr_table.put(id.getText(), r);		
				} 
				else if (type.getText().equals("boolean")) 
				{
					r = new VinlBoolean(id.getText());
					curr_table.put(id.getText(), r);		
				} 
				else 
				{
					throw new RecognitionException("ERROR: Unrecognized type specified.");
				}
			}
		}
	| var_id:ID
		{
			if(proceed) 
			{
				if(curr_table.get(var_id.getText())!=null)  
					r=curr_table.get(var_id.getText());
				else
					throw new RecognitionException("ERROR: Undeclared variable: "+var_id.getText());
			}
		}
	| decimal:DECIMAL		{ if(proceed) r=new VinlDecimal(Double.parseDouble(decimal.getText())); }
	| #("echo" a=expr) 		
		{ 
			if(proceed)
			{
				String value="";
				if(a instanceof VinlBoolean) 
					value += ((VinlBoolean)a).getValue();
				else if(a instanceof VinlDecimal) 
					value += ((VinlDecimal)a).getValue();
				else if(a instanceof VinlFunction) 
					value += "a function";
				else
					value="not valid"; 
				log("The rvalue is: "+value, curr_level+1); //always print 
			}
		}
	| #("forward" a=expr) 		
		{ 
			if(proceed) 
			{
				log("Going forward: "+((VinlDecimal)a).getValue(), DEBUG_FINE); 
				gp.forward(((VinlDecimal)a).getValue());
			}
		}
	| #("fw" a=expr) 		
		{ 
			if(proceed) 
			{
				log("Going forward: "+((VinlDecimal)a).getValue(), DEBUG_FINE); 
				gp.forward(((VinlDecimal)a).getValue());
			}
		}
	| #("backward" a=expr) 		
		{ 
			if(proceed)
			{
				log("Going backward: "+((VinlDecimal)a).getValue(), DEBUG_FINE); 
				gp.backward(((VinlDecimal)a).getValue());
			}
		}
	| #("bw" a=expr) 		
		{ 
			if(proceed)
			{
				log("Going backward: "+((VinlDecimal)a).getValue(), DEBUG_FINE); 
				gp.backward(((VinlDecimal)a).getValue());
			}
		}
	| #("left" a=expr) 		
		{ 
			if(proceed)
			{
				log("Turn left: "+((VinlDecimal)a).getValue(), DEBUG_FINE); 
				gp.left(((VinlDecimal)a).getValue());
			}
		}
	| #("lf" a=expr) 		
		{ 
			if(proceed)
			{
				log("Turn left: "+((VinlDecimal)a).getValue(), DEBUG_FINE); 
				gp.left(((VinlDecimal)a).getValue());
			}
		}
	| #("right" a=expr) 		
		{ 
			if(proceed)
			{
				log("Turn right: "+((VinlDecimal)a).getValue(), DEBUG_FINE); 
				gp.right(((VinlDecimal)a).getValue());
			}
		}
	| #("rt" a=expr) 		
		{ 
			if(proceed)
			{
				log("Turn right: "+((VinlDecimal)a).getValue(), DEBUG_FINE); 
				gp.right(((VinlDecimal)a).getValue());
			}
		}
	| #("if" a=expr ifbody:. (elsebody:.)?) 
		{
			if(proceed)
			{
				if(a instanceof VinlBoolean) 
				{
					if(((VinlBoolean)a).getValue()) { 
						//do if part
						r = expr(#ifbody);
					}
					else 
					{
						//do else part
						r = expr(#elsebody);
					}
				} 
				else
				{
					throw new RecognitionException("ERROR: Only accept boolean express as if condition");
				} 
			}
		}
	| #("return" 
		( 
			a=expr { if(proceed) { ret = a; r = a; } }
		)?
	   ) 
		{
			/* if returned, it's same as breaked in the sense no stmt should run anymore */
			proceed = false;
			returned = true;
		}
	| #("while" while_condition:. while_body:.) 
		{
			if(proceed) 
			{
				inloop = true;
				VinlType while_cond = expr(#while_condition);	
				if(!(while_cond instanceof VinlBoolean)) 
					throw new RecognitionException("ERROR: Condition in for/while loop must be boolean type");
				while(((VinlBoolean)while_cond).getValue()) 
				{
					r = expr(#while_body);
					if(breaked) 
					{
						//reset break for future statements
						proceed = true;
						breaked = false;
						break;
					}
					else if(continued) 
					{
						proceed=true;
						continued=false;
					}
					while_cond=expr(#while_condition);
				}
				inloop = false;	
			}
		}
	| #("for" a=expr for_condition:.  post_action:. for_body:.) 
		{
			if(proceed) 
			{
				inloop = true;
				VinlType for_cond = expr(#for_condition);
				if(!(for_cond instanceof VinlBoolean))
					throw new RecognitionException("ERROR: Condition in for/while loop must be boolean type");
				while(((VinlBoolean)for_cond).getValue()) 
				{
					r = expr(#for_body);
					if(breaked) 
					{
						//reset break for future statements
						proceed = true;
						breaked = false;
						break;
					}
					else if(continued) 
					{
						proceed=true;
						continued=false;
					}
					r = expr(#post_action);
					for_cond=expr(#for_condition);
				}
				inloop = false;	
			}	
		}
	| "status" 
		{
			/* printout internal states*/
			log(String.format("Control flows: proceed=(%s), inloop=(%s), breaked=(%s), continued=(%s), returned=(%s)", proceed, inloop, breaked, continued, returned), curr_level+1);
			log(String.format("turtle coordinates: x=%f y=%f angle(in degree)=%f", turtle.getX(), turtle.getY(), turtle.getAngle()), curr_level+1);
		}
	| #(FUNC_CALL a=expr expr_list=mexpr) 
		{
			if(proceed) {
				if(!(a instanceof VinlFunction))
					throw new RecognitionException("ERROR: "+a.getName()+" is not a function.");
				
				VinlFunction func = (VinlFunction) a;
				/* every function starts its own scope with current scope as it's parent scope */
				log("Current symbol table: " + curr_table, DEBUG_FINE);
				func.getSymTable().setParent(curr_table);
				curr_table = func.getSymTable();
				log("New symbol table, going to functions: "+func.getName()+"'s scope: " + curr_table, DEBUG_FINE);
				/* evaluate arguments and put them in symbol table */	
				String[] params = func.getParams();
				int j = 0; //expr_list index
				for(int i = 0; i<params.length; i++) {
					if(params[i].equals("void")) {
						if(expr_list[j]==null || !(expr_list[j] instanceof VinlVoid)) 
							throw new RecognitionException("ERROR: Wrong argument passed to function at index: "+j);
					} else if (params[i].equals("decimal") && params[++i] != null) {
						if(expr_list[j]==null || !(expr_list[j] instanceof VinlDecimal)) 
							throw new RecognitionException("ERROR: Wrong argument passed to function at index: "+j);
						log("putting "+params[i]+" in sym table",DEBUG_FINE);
						((VinlDecimal)expr_list[j]).setName(params[i]);
						curr_table.put(params[i], expr_list[j++]); 
					} else if (params[i].equals("boolean") && params[++i] != null) {
						if(expr_list[j]==null || !(expr_list[j] instanceof VinlBoolean)) 
							throw new RecognitionException("ERROR: Wrong argument passed to function at index: "+j);
						log("putting "+params[i]+" in sym table",DEBUG_FINE);
						((VinlBoolean)expr_list[j]).setName(params[i]);
						curr_table.put(params[i], expr_list[j++]); 
					}
				}
				/* invoke this function */	
				r=expr(func.getBody());
				if(returned) {
					proceed = true; 
					returned = false;
					if(ret != null) {
						r = ret;
					}
					ret = null;
				}
				/* coming out of the function so resume parent scope now */
				log("Out of function: "+func.getName()+"'s scope: " + curr_table, DEBUG_FINE);
				curr_table = curr_table.getParent();
				log("Back to "+func.getName()+"'s parent scope: " + curr_table, DEBUG_FINE);
				/* destory the function symbol table */
				func.setSymTable(new VinlSymbolTable());
			}
		}
	| #("funct" func_name:ID func_params=vlist func_body:.)
		{
			r = new VinlFunction(func_name.getText(), func_params, #func_body);
			curr_table.put(func_name.getText(), r);
		}
	;

mexpr returns [ VinlType[] mret ]
{
	ArrayList<VinlType> list = new ArrayList();
	mret = null;
	VinlType a;
}
	: #(EXPR_LIST	(a=expr {list.add(a);})+) 
		{
			mret = new VinlType[list.size()];
			mret = list.toArray(mret); 
		}
	;

vlist returns [ String[] ret ] 
{
	/* type and variable name alternating unless it's void*/
	ArrayList<String> list=new ArrayList();
	ret = null;
} 
	: #(PARAMS_LIST	(s:.	{list.add(s.getText());})+) 
		{
			for(int i = 0; i < list.size(); i++){
				log(list.get(i), DEBUG_FINE);	
			} 
			ret = new String[list.size()]; 
			ret = list.toArray(ret); 
		}
	;

